////////////////////////////////////////////////////////////////////////////////
//
//
//  FileName    :   ssSimpleIPC.pas
//  Creator     :   Shen Min
//  Date        :   2003-04-05
//  Comment     :
//
//  Copyright (c) 2002-2003 Sunisoft
//  http://www.sunisoft.com
//  Email: support@sunisoft.com
//
////////////////////////////////////////////////////////////////////////////////

unit ssSimpleIPC;

interface

uses Windows, Messages, Classes, SysUtils;

type
    TssWndObjRec = record
        WndHandle : HWND;
        pObject : TObject;
    end;

    ssSimpleIPC_WndProc = function (hwnd : HWND; uMsg : Cardinal; wParam : WPARAM; lParam : LPARAM) : LResult; stdcall;

    TssIPCTextMsgRecieved = procedure (Sender : TObject; MsgBuf : String) of object;
    TssIPCBinMsgRecieved = procedure (Sender : TObject; MsgBuf : Pointer; MsgSize : Integer) of object;

    TssSimpleIPC = class(TComponent)
    private
        m_Name : String;

        m_OnTextMsgRecieved : TssIPCTextMsgRecieved;
        m_OnBinMsgRecieved : TssIPCBinMsgRecieved;

        m_WndObj : TssWndObjRec;

        procedure SetIPCName(const Value: String);

        procedure TextMsgRecieved(MsgBuf : String);
        procedure BinMsgRecieved(MsgBuf : Pointer; MsgSize : Integer);

    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy(); override;

        procedure SendTextMessage(StrSend : String);
        procedure SendBinMessage(pBuffer : Pointer; nBufSize : Integer);

    published
        property IPCName : String read m_Name write SetIPCName;

        property OnTextMsgRecieved : TssIPCTextMsgRecieved read m_OnTextMsgRecieved write m_OnTextMsgRecieved;
        property OnBinMsgRecieved : TssIPCBinMsgRecieved read m_OnBinMsgRecieved write m_OnBinMsgRecieved;

    end;

    procedure RaiseOSError();

    function CreateCommWnd(WndName : String; WndProc : ssSimpleIPC_WndProc) : HWND;
    procedure DestroyCommWnd(WndHandle : HWND);

    procedure Register;

implementation


procedure Register;
begin
    RegisterComponents('Sunisoft FreeVCL', [TssSimpleIPC]);
end;

var
    l_lstObject : TList;

function GetObjectFromWnd(WndHandle : HWND) : TssSimpleIPC;
var
    i : Integer;
    WndObj : TssWndObjRec;
begin
    Result := nil;

    for i := 0 to l_lstObject.Count - 1 do
    begin
        WndObj := TssWndObjRec(l_lstObject[i]^);
        if WndObj.WndHandle = WndHandle then
        begin
            Result := WndObj.pObject as TssSimpleIPC;
            break;
        end;
    end;
end;

function CommWndProc(hwnd : HWND; uMsg : Cardinal; wParam : WPARAM; lParam : LPARAM) : LResult; stdcall;
var
    Obj : TssSimpleIPC;
    pCopyData : PCOPYDATASTRUCT;
begin
    Result := 0;

    if uMsg = WM_COPYDATA then
    begin
        Obj := GetObjectFromWnd(hwnd);
        if Obj <> nil then
        begin
            pCopyData := PCOPYDATASTRUCT(lParam);
            if pCopyData.dwData = 0 then
                Obj.BinMsgRecieved(pCopyData.lpData, pCopyData.cbData)
            else
                Obj.TextMsgRecieved(Copy(PChar(pCopyData.lpData), 0, pCopyData.cbData - 1));
        end;
    end
    else
        Result := DefWindowProc(hWnd, uMsg, wParam, lParam);
end;

{ TssSimpleIPC }

procedure TssSimpleIPC.BinMsgRecieved(MsgBuf: Pointer; MsgSize: Integer);
begin
    if Assigned(m_OnBinMsgRecieved) then
        m_OnBinMsgRecieved(self, MsgBuf, MsgSize);
end;

constructor TssSimpleIPC.Create(AOwner: TComponent);
var
    CommWnd : HWND;
begin
    inherited;

    IPCName := 'ssSimpleIPC_IPCName';

    if not (csDesigning in ComponentState) then
    begin
        CommWnd := CreateCommWnd(IPCName, CommWndProc);
        if CommWnd = 0 then
            raise Exception.Create('DataCopyIPC object create failed');

        m_WndObj.WndHandle := CommWnd;
        m_WndObj.pObject := self;
        l_lstObject.Add(@m_WndObj);
    end;
end;

destructor TssSimpleIPC.Destroy;
begin
    if not (csDesigning in ComponentState) then
    begin
        l_lstObject.Remove(@m_WndObj);
        DestroyCommWnd(m_WndObj.WndHandle);
    end;

    inherited;
end;

procedure TssSimpleIPC.SendBinMessage(pBuffer: Pointer; nBufSize : Integer);
var
    CopyData : COPYDATASTRUCT;
    DestWnd : HWND;
begin
    CopyData.dwData := 0; // binary
    CopyData.cbData := nBufSize;
    CopyData.lpData := pBuffer;

    DestWnd := FindWindowEx(0, 0, nil, PChar(IPCName));
    while DestWnd <> 0 do
    begin
        if DestWnd <> m_WndObj.WndHandle then
            SendMessage(DestWnd, WM_COPYDATA, m_WndObj.WndHandle, Integer(@CopyData));
        DestWnd := FindWindowEx(0, DestWnd, nil, PChar(IPCName));
    end;
end;

procedure TssSimpleIPC.SendTextMessage(StrSend: String);
var
    CopyData : COPYDATASTRUCT;
    DestWnd : HWND;
begin
    CopyData.dwData := 1; // string
    CopyData.cbData := Length(StrSend) + 1;
    CopyData.lpData := PChar(StrSend);

    DestWnd := FindWindowEx(0, 0, nil, PChar(IPCName));
    while DestWnd <> 0 do
    begin
        if DestWnd <> m_WndObj.WndHandle then
            SendMessage(DestWnd, WM_COPYDATA, m_WndObj.WndHandle, Integer(@CopyData));
        DestWnd := FindWindowEx(0, DestWnd, nil, PChar(IPCName));
    end;
end;

procedure TssSimpleIPC.SetIPCName(const Value: String);
begin
    if m_Name <> Value then
    begin
        m_Name := Value;

        if m_WndObj.WndHandle = 0 then
            Exit;
        SetWindowText(m_WndObj.WndHandle, PChar(IPCName));
    end;
end;

procedure TssSimpleIPC.TextMsgRecieved(MsgBuf: String);
begin
    if Assigned(m_OnTextMsgRecieved) then
        m_OnTextMsgRecieved(self, MsgBuf);
end;

{ Functions }

procedure RaiseOSError();
begin
    RaiseLastOSError(); //RaiseLastWin32Error();
end;

function CreateCommWnd(WndName : String; WndProc : ssSimpleIPC_WndProc) : HWND;
var
    wndcls : WNDCLASS;
    hWnd : THandle;
begin
    Result := 0;

    wndcls.style         := 0;
    wndcls.lpfnWndProc   := @WndProc;
    wndcls.cbClsExtra    := 0;
    wndcls.cbWndExtra    := 0;
    wndcls.hInstance     := hInstance;
    wndcls.hIcon         := 0;
    wndcls.hCursor       := 0;
    wndcls.hbrBackground := 0;
    wndcls.lpszMenuName  := nil;
    wndcls.lpszClassName := 'ssSimpleIPC_CommWnd';
    if Windows.RegisterClass(wndcls) = 0 then
    begin
        if GetLastError() <> 1410 then
            Exit;
    end;

    hWnd := CreateWindow(
        'ssSimpleIPC_CommWnd',
        PChar(WndName),
        WS_BORDER,
        Integer(0),
        Integer(0),
        Integer(0),
        Integer(0),
        0,
        0,
        hInstance,
        nil
    );

    Result := hWnd;
end;

procedure DestroyCommWnd(WndHandle : HWND);
begin
    DestroyWindow(WndHandle);
end;

initialization
    l_lstObject := TList.Create();

finalization
    l_lstObject.Free();

end.
